### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2010 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages

# import Zope3 interfaces
from zope.traversing.interfaces import TraversalError

# import local interfaces
from hurry.workflow.interfaces import IWorkflowState
from ztfy.workflow.interfaces import IWorkflow, IWorkflowTarget

# import Zope3 packages
from zope.component import getUtility, queryMultiAdapter
from zope.interface import Interface
from zope.traversing import namespace

# import local packages


class WorkflowNamespaceTraverser(namespace.view):
    """Workflow namespace traverser"""

    def traverse(self, name, ignored):
        try:
            workflow = getUtility(IWorkflow, IWorkflowTarget(self.context).workflow_name)
            state = IWorkflowState(self.context).getState()
            transition = workflow.getTransition(state, name)  # raise InvalidTransactionError if NOK
            view_name = transition.user_data.get('view')
            if view_name is not None:
                view = queryMultiAdapter((self.context, self.request), Interface, view_name)
                if view is not None:
                    view.transition = transition
                    return view
        except:
            pass
        raise TraversalError("++wf++%s" % name)



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2009 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages
import pytz
from datetime import datetime

# import Zope3 interfaces

# import local interfaces
from hurry.workflow.interfaces import IWorkflowTransitionEvent
from ztfy.comment.interfaces import IComments
from ztfy.utils.request import getRequest
from ztfy.workflow.interfaces import IWorkflow, IWorkflowTarget, IWorkflowContent

# import Zope3 packages
from zope.component import adapter, getUtility
from zope.i18n import translate

# import local packages

from ztfy.workflow import _


@adapter(IWorkflowTarget, IWorkflowTransitionEvent)
def handleWorkflowTransition(object, event):
    request = getRequest()
    content = IWorkflowContent(object, None)
    if content is not None:
        content.state_date = datetime.now(pytz.UTC)
        content.state_principal = request.principal.id
    wf = getUtility(IWorkflow, IWorkflowTarget(object).workflow_name)
    if event.source is not None:
        comment = translate(_('Changed state: from %s to %s'), context=request) % (translate(wf.states.getTerm(event.source).title, context=request),
                                                                                   translate(wf.states.getTerm(event.destination).title, context=request))
    else:
        comment = translate(_('New state: %s'), context=request) % (translate(wf.states.getTerm(event.destination).title, context=request))
    if event.comment:
        comment += '\n%s' % event.comment
    comments = IComments(object)
    comments.addComment(body=comment,
                        renderer='zope.source.plaintext', tags=('__workflow__'))



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2009 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages

# import Zope3 interfaces
from ztfy.workflow.browser.viewlet.interfaces import IWorkflowViewletManager, IWorkflowTransitionsViewlet

# import local interfaces
from hurry.workflow.interfaces import IWorkflowInfo
from ztfy.workflow.interfaces import IWorkflow, IWorkflowTarget, ITransitionTarget

# import Zope3 packages
from z3c.template.template import getPageTemplate
from zope.component import queryUtility, queryMultiAdapter
from zope.interface import implements
from zope.viewlet.manager import WeightOrderedViewletManager
from zope.viewlet.viewlet import ViewletBase

# import local packages


class WorkflowViewletManager(WeightOrderedViewletManager):

    implements(IWorkflowViewletManager)


class WorkflowTransitionsViewlet(ViewletBase):

    implements(IWorkflowTransitionsViewlet)

    def _getTransitions(self):
        _transitions = []
        target = IWorkflowTarget(self.context, None)
        if target is None:
            return _transitions
        wf = queryUtility(IWorkflow, target.workflow_name)
        if wf is None:
            return _transitions
        info = IWorkflowInfo(self.context)
        ids = info.getManualTransitionIds()
        return [t for t in (wf.getTransitionById(id) for id in ids)
                        if t.user_data.get('view')]

    def update(self):
        self.transitions = self._getTransitions()

    render = getPageTemplate()

    def getURL(self, transition):
        target = queryMultiAdapter((self.context, self.request, transition), ITransitionTarget)
        if target is not None:
            return target.absoluteURL()
        else:
            return transition.user_data.get('view')



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2009 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import Zope3 interfaces
from zope.security.interfaces import NoInteraction, Unauthorized

# import local interfaces
from hurry.workflow.interfaces import IWorkflowInfo, IWorkflowState, ConditionFailedError
from ztfy.workflow.interfaces import IWorkflow, IWorkflowTarget, ITransition, ITransitionTarget

# import Zope3 packages
from zope.component import adapts, getUtility
from zope.event import notify
from zope.interface import Interface, implements
from zope.lifecycleevent import ObjectModifiedEvent
from zope.security.management import getInteraction
from zope.traversing.browser import absoluteURL

# import local packages
from hurry.workflow.workflow import nullCheckPermission, WorkflowTransitionEvent, WorkflowVersionTransitionEvent
from hurry.workflow.workflow import Workflow as WorkflowBase
from hurry.workflow.workflow import WorkflowInfo as WorkflowInfoBase, WorkflowState as WorkflowStateBase


class WorkflowStateAdapter(WorkflowStateBase):
    """Workflow state adapter implementing ILocation"""

    adapts(IWorkflowTarget)
    implements(IWorkflowState)


class WorkflowInfoAdapter(WorkflowInfoBase):
    """Enhanced IWorkflowInfo adapter handling several registered workflows"""

    adapts(IWorkflowTarget)
    implements(IWorkflowInfo)

    @property
    def name(self):
        return IWorkflowTarget(self.context).workflow_name

    def fireTransition(self, transition_id, comment=None, side_effect=None,
                       check_security=True):
        state = IWorkflowState(self.context)
        wf = getUtility(IWorkflow, self.name)
        # this raises InvalidTransitionError if id is invalid for current state
        transition = wf.getTransition(state.getState(), transition_id)
        # check whether we may execute this workflow transition
        try:
            interaction = getInteraction()
        except NoInteraction:
            checkPermission = nullCheckPermission
        else:
            if check_security:
                checkPermission = interaction.checkPermission
            else:
                checkPermission = nullCheckPermission
        if not checkPermission(
            transition.permission, self.context):
            raise Unauthorized(self.context,
                               'transition: %s' % transition_id,
                               transition.permission)
        # now make sure transition can still work in this context
        if not transition.condition(self, self.context):
            raise ConditionFailedError
        # perform action, return any result as new version
        result = transition.action(self, self.context)
        if result is not None:
            if transition.source is None:
                IWorkflowState(result).initialize()
            # stamp it with version
            state = IWorkflowState(result)
            state.setId(IWorkflowState(self.context).getId())
            # execute any side effect:
            if side_effect is not None:
                side_effect(result)
            event = WorkflowVersionTransitionEvent(result, self.context,
                                                   transition.source,
                                                   transition.destination,
                                                   transition, comment)
        else:
            if transition.source is None:
                IWorkflowState(self.context).initialize()
            # execute any side effect
            if side_effect is not None:
                side_effect(self.context)
            event = WorkflowTransitionEvent(self.context,
                                            transition.source,
                                            transition.destination,
                                            transition, comment)
        # change state of context or new object
        state.setState(transition.destination)
        notify(event)
        # send modified event for original or new object
        if result is None:
            notify(ObjectModifiedEvent(self.context))
        else:
            notify(ObjectModifiedEvent(result))
        return result

    def getFireableTransitionIdsToward(self, state):
        wf = getUtility(IWorkflow, self.name)
        result = []
        for transition_id in self.getFireableTransitionIds():
            transition = wf.getTransitionById(transition_id)
            if transition.destination == state:
                result.append(transition_id)
        return result

    def _getTransitions(self, trigger):
        # retrieve all possible transitions from workflow utility
        wf = getUtility(IWorkflow, self.name)
        transitions = wf.getTransitions(IWorkflowState(self.context).getState())
        # now filter these transitions to retrieve all possible
        # transitions in this context, and return their ids
        return [transition for transition in transitions
                                          if transition.trigger == trigger]


class Workflow(WorkflowBase):
    """Custom workflow class"""

    implements(IWorkflow)

    def __init__(self, transitions, states, published_states=()):
        super(Workflow, self).__init__(transitions)
        self.states = states
        self.published_states = published_states


class WorkflowTransitionTargetAdapter(object):

    adapts(Interface, Interface, ITransition)
    implements(ITransitionTarget)

    def __init__(self, context, request, transition):
        self.context = context
        self.request = request
        self.transition = transition

    def absoluteURL(self):
        return '%s/++wf++%s' % (absoluteURL(self.context, self.request),
                                self.transition.transition_id)



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2011 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages
import transaction

# import Zope3 interfaces
from zope.app.publication.zopepublication import ZopePublication
from zope.catalog.interfaces import ICatalog
from zope.component.interfaces import IComponentRegistry, ISite
from zope.dublincore.interfaces import IZopeDublinCore
from zope.intid.interfaces import IIntIds
from zope.processlifetime import IDatabaseOpenedWithRoot

# import local interfaces
from hurry.workflow.interfaces import IWorkflowState
from ztfy.utils.interfaces import INewSiteManagerEvent

# import Zope3 packages
from zc.catalog.catalogindex import ValueIndex, DateTimeValueIndex
from zope.catalog.catalog import Catalog
from zope.component import adapter, queryUtility
from zope.intid import IntIds
from zope.location import locate
from zope.site import hooks

# import local packages
from ztfy.utils.site import locateAndRegister


def updateDatabaseIfNeeded(context):
    """Check for missing objects at application startup"""
    try:
        sm = context.getSiteManager()
    except:
        return
    default = sm['default']
    # Check for required IIntIds utility
    intids = queryUtility(IIntIds)
    if intids is None:
        intids = default.get('IntIds')
        if intids is None:
            intids = IntIds()
            locate(intids, default)
            default['IntIds'] = intids
            IComponentRegistry(sm).registerUtility(intids, IIntIds)
    # Check for required catalog and index
    catalog = default.get('WorkflowCatalog')
    if catalog is None:
        catalog = Catalog()
        locateAndRegister(catalog, default, 'WorkflowCatalog', intids)
        IComponentRegistry(sm).registerUtility(catalog, ICatalog, 'WorkflowCatalog')
    if catalog is not None:
        if 'wf_id' not in catalog:
            index = ValueIndex('getId', IWorkflowState, True)
            locateAndRegister(index, catalog, 'wf_id', intids)
        if 'wf_state' not in catalog:
            index = ValueIndex('getState', IWorkflowState, True)
            locateAndRegister(index, catalog, 'wf_state', intids)
        if 'creation_date' not in catalog:
            index = DateTimeValueIndex('created', IZopeDublinCore, False)
            locateAndRegister(index, catalog, 'creation_date', intids)
        if 'modification_date' not in catalog:
            index = DateTimeValueIndex('modified', IZopeDublinCore, False)
            locateAndRegister(index, catalog, 'modification_date', intids)
        if 'effective_date' not in catalog:
            index = DateTimeValueIndex('effective', IZopeDublinCore, False)
            locateAndRegister(index, catalog, 'effective_date', intids)
        if 'expiration_date' not in catalog:
            index = DateTimeValueIndex('expires', IZopeDublinCore, False)
            locateAndRegister(index, catalog, 'expiration_date', intids)


@adapter(IDatabaseOpenedWithRoot)
def handleOpenedDatabase(event):
    db = event.database
    connection = db.open()
    root = connection.root()
    root_folder = root.get(ZopePublication.root_name, None)
    for site in root_folder.values():
        if ISite(site, None) is not None:
            hooks.setSite(site)
            updateDatabaseIfNeeded(site)
            transaction.commit()


@adapter(INewSiteManagerEvent)
def handleNewSiteManager(event):
    updateDatabaseIfNeeded(event.object)



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2009 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


from datetime import datetime
from persistent.dict import PersistentDict
import pytz

# import Zope3 interfaces
from zope.annotation.interfaces import IAnnotations
from zope.dublincore.interfaces import IZopeDublinCore

# import local interfaces

# import Zope3 packages
from zope.component import adapts, queryUtility
from zope.interface import implements

# import local packages
from hurry.workflow.interfaces import IWorkflowState
from ztfy.security.security import getSecurityManager
from ztfy.utils.date import unidate, parsedate
from ztfy.utils.traversing import getParent
from ztfy.workflow.interfaces import IWorkflow, IWorkflowTarget, IWorkflowContent


WORKFLOW_CONTENT_KEY = 'ztfy.workflow.content'

GMT = pytz.timezone('GMT')


class WorkflowContentAdapter(object):

    adapts(IWorkflowTarget)
    implements(IWorkflowContent)

    def __init__(self, context):
        self.context = context
        annotations = IAnnotations(self.context)
        data = annotations.get(WORKFLOW_CONTENT_KEY)
        if data is None:
            data = annotations[WORKFLOW_CONTENT_KEY] = PersistentDict()
        self.data = data

    def _getStateDate(self):
        return parsedate(self.data.get('state_date')) or IZopeDublinCore(self.context).created

    def _setStateDate(self, date):
        if date and not date.tzinfo:
            date = GMT.localize(date)
        self.data['state_date'] = unidate(date)

    state_date = property(_getStateDate, _setStateDate)

    def _getStatePrincipal(self):
        return self.data.get('state_principal') or IZopeDublinCore(self.context).creators[0]

    def _setStatePrincipal(self, principal):
        self.data['state_principal'] = principal

    state_principal = property(_getStatePrincipal, _setStatePrincipal)

    def _getPublicationDate(self):
        return parsedate(self.data.get('publication_date'))

    def _setPublicationDate(self, date):
        if date and not date.tzinfo:
            date = GMT.localize(date)
        self.data['publication_date'] = unidate(date)

    publication_date = property(_getPublicationDate, _setPublicationDate)

    def _getFirstPublicationDate(self):
        return parsedate(self.data.get('first_publication_date'))

    def _setFirstPublicationDate(self, date):
        if date and not date.tzinfo:
            date = GMT.localize(date)
        self.data['first_publication_date'] = unidate(date)

    first_publication_date = property(_getFirstPublicationDate, _setFirstPublicationDate)

    def _getEffectiveDate(self):
        return IZopeDublinCore(self.context).effective

    def _setEffectiveDate(self, date):
        if date:
            if not date.tzinfo:
                date = GMT.localize(date)
            IZopeDublinCore(self.context).effective = date
            if (self.first_publication_date is None) or (self.first_publication_date > date):
                self.first_publication_date = date
        else:
            if self.first_publication_date == self.publication_effective_date:
                self.first_publication_date = None
            if self.publication_effective_date:
                del IZopeDublinCore(self.context)._mapping['Date.Effective']

    publication_effective_date = property(_getEffectiveDate, _setEffectiveDate)

    def _getExpirationDate(self):
        return IZopeDublinCore(self.context).expires

    def _setExpirationDate(self, date):
        if date:
            if not date.tzinfo:
                date = GMT.localize(date)
            IZopeDublinCore(self.context).expires = date
        else:
            if self.publication_expiration_date:
                del IZopeDublinCore(self.context)._mapping['Date.Expires']

    publication_expiration_date = property(_getExpirationDate, _setExpirationDate)

    def isPublished(self):
        parent = getParent(self.context, IWorkflowTarget, allow_context=False)
        if (parent is not None) and (not IWorkflowContent(parent).isPublished()):
            return False
        workflow_name = IWorkflowTarget(self.context).workflow_name
        if not workflow_name:
            return True
        workflow = queryUtility(IWorkflow, workflow_name)
        if (workflow is None) or (not workflow.published_states):
            return False
        if IWorkflowState(self.context).getState() not in workflow.published_states:
            return False
        now = datetime.now(pytz.UTC)
        return (self.publication_effective_date is not None) and \
               (self.publication_effective_date <= now) and \
               ((self.publication_expiration_date is None) or \
                (self.publication_expiration_date >= now))

    def isVisible(self):
        result = True
        sm = getSecurityManager(self.context)
        if sm is not None:
            result = sm.canView()
        return result and self.isPublished()



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2010 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages

# import Zope3 interfaces
from zope.tales.interfaces import ITALESFunctionNamespace

# import local interfaces
from hurry.workflow.interfaces import IWorkflowState
from ztfy.workflow.interfaces import IWorkflow, IWorkflowTarget
from ztfy.workflow.tal.interfaces import IWorkflowTalesAPI

# import Zope3 packages
from zope.component import queryUtility
from zope.i18n import translate
from zope.interface import implements

# import local packages

from ztfy.workflow import _


class WorkflowTalesAdapter(object):

    implements(IWorkflowTalesAPI, ITALESFunctionNamespace)

    def __init__(self, context):
        self.context = context

    def setEngine(self, engine):
        self.request = engine.vars['request']

    def status(self):
        target = IWorkflowTarget(self.context, None)
        if target is None:
            return translate(_("None"), context=self.request)
        wf = queryUtility(IWorkflow, target.workflow_name)
        if wf is None:
            return translate(_("None"), context=self.request)
        state = IWorkflowState(self.context).getState()
        return translate(wf.states.getTerm(state).title, context=self.request)
